from eppy.iddcurrent import iddcurrent
import pytest
from six import StringIO

from geomeppy.idf import IDF
from geomeppy.extractor import copy_constructions, copy_geometry

idf_txt = """
Version, 8.8;
Zone, z1 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
Material, Spam, Rough, 0.1, 1, 1000, 1000, 0.9, 0.9, 0.9;
"""

@pytest.fixture()
def base_idf():
    # type: () -> None
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)

    return IDF(StringIO(idf_txt))


def test_copy_geometry(base_idf):
    idf = copy_geometry(base_idf)
    assert idf.getobject('ZONE', 'z1 Thermal Zone')
    assert not idf.getobject('MATERIAL', 'Spam')


def test_copy_constructions(base_idf):
    idf = copy_constructions(base_idf)
    assert not idf.getobject('ZONE', 'z1 Thermal Zone')
    assert idf.getobject('MATERIAL', 'Spam')



"""Tests for recipes."""
from eppy.iddcurrent import iddcurrent
import pytest
from six import StringIO

from geomeppy.idf import IDF
from geomeppy.geom.intersect_match import intersect_idf_surfaces, match_idf_surfaces
from geomeppy.geom.polygons import Polygon3D
from geomeppy.geom.vectors import Vector2D, Vector3D
from geomeppy.recipes import rotate, set_wwr, translate, translate_to_origin
from geomeppy.utilities import almostequal
from geomeppy.view_geometry import _get_collections, _get_shading, _get_surfaces


idf_txt = """
Version, 8.5;
Building, Building 1, , , , , , , ;
Zone, z1 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
Zone, z2 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
BuildingSurface:Detailed, z1_FLOOR, Floor, , z1 Thermal Zone, ground, , NoSun, NoWind, , , 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0;
BuildingSurface:Detailed, z1_ROOF, Roof, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 2.0, 0.5, 1.0, 2.1, 0.5, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0001, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.0, 1.1, 0.5, 1.0, 1.1, 0.0, 1.0, 2.1, 0.0, 1.0, 2.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0002, Wall, , z1 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 1.0, 2.1, 0.5, 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 2.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0003, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 2.0, 0.5, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 2.0, 1.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0004, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z2_FLOOR, Floor, , z2 Thermal Zone, ground, , NoSun, NoWind, , , 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0;
BuildingSurface:Detailed, z2_ROOF, Roof, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 2.95, 0.5, 1.5, 3.05, 0.5, 1.5, 2.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0001, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 2.05, 0.5, 1.5, 2.05, 0.0, 1.5, 3.05, 0.0, 1.5, 3.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0002, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 3.05, 0.5, 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 2.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0003, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 2.95, 0.5, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 2.5, 1.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0004, Wall, , z2 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0, 1.5, 2.05, 0.5;
Shading:Zone:Detailed, z2_SHADE_0003, z2_WALL_0003, , 4, 2.5, 2.95, 0.5, 2.6, 2.95, 0.3, 2.6, 1.95, 0.3, 2.5, 1.95, 0.5;
"""

@pytest.fixture()
def base_idf():
    # type: () -> None
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)

    return IDF(StringIO(idf_txt))


class TestTranslate():
    
    def test_translate(self, base_idf):
        # type: () -> None
        idf = base_idf
        surfaces = idf.getsurfaces()
        translate(surfaces, (50, 100))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected
        translate(surfaces, (-50, -100))  # move back
        
        translate(surfaces, Vector2D(50, 100))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected
        translate(surfaces, (-50, -100))  # move back
        
        translate(surfaces, Vector3D(50, 100, 0))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [52.0, 52.0, 51.0, 51.0]
        assert result == expected

    def test_translate_to_origin(self, base_idf):
        # type: () -> None
        idf = base_idf
        surfaces = idf.getsurfaces()
        translate(surfaces, (50000, 10000))  # move to x + 50, y + 100
        result = Polygon3D(surfaces[0].coords).xs
        expected = [50002.0, 50002.0, 50001.0, 50001.0]
        assert result == expected

        translate_to_origin(idf)
        surfaces = idf.getsurfaces()
        min_x = min(min(Polygon3D(s.coords).xs) for s in surfaces)
        min_y = min(min(Polygon3D(s.coords).ys) for s in surfaces)
        assert min_x == 0
        assert min_y == 0

    def test_rotate_360(self, base_idf):
        # type: () -> None
        idf = base_idf
        surface = idf.getsurfaces()[0]
        coords = [Vector3D(*v) for v in [(0,0,0), (1,0,0), (1,1,0), (0,1,0)]]
        surface.setcoords(coords)
        expected = surface.coords
        rotate([surface], 360)
        result = surface.coords
        assert almostequal(result, expected)

    def test_rotate_idf_360(self, base_idf):
        # type: () -> None
        idf1 = base_idf
        idf2 = IDF()
        idf2.initreadtxt(idf1.idfstr())
        idf1.rotate(360)
        floor1 = Polygon3D(idf1.getsurfaces('floor')[0].coords).normalize_coords(None)
        floor2 = Polygon3D(idf2.getsurfaces('floor')[0].coords).normalize_coords(None)
        assert almostequal(floor1, floor2)
        shade1 = Polygon3D(idf1.getshadingsurfaces()[0].coords).normalize_coords(None)
        shade2 = Polygon3D(idf1.getshadingsurfaces()[0].coords).normalize_coords(None)
        assert almostequal(shade1, shade2)

    def test_centre(self, base_idf):
        # type: () -> None
        idf = base_idf
        result = idf.centroid
        expected = Vector2D(1.75, 2.025)
        assert result == expected

    def test_scale_idf(self, base_idf):
        # type: () -> None
        idf1 = base_idf
        idf2 = IDF()
        idf2.initreadtxt(idf1.idfstr())
        idf1.scale(10)
        idf1.scale(0.1)
        floor1 = Polygon3D(idf1.getsurfaces('floor')[0].coords).normalize_coords(None)
        floor2 = Polygon3D(idf2.getsurfaces('floor')[0].coords).normalize_coords(None)
        assert almostequal(floor1, floor2)


class TestMatchSurfaces():

    def test_set_wwr(self, base_idf):
        # type: () -> None
        idf = base_idf
        intersect_idf_surfaces(idf)
        match_idf_surfaces(idf)
        wwr = 0.25
        set_wwr(idf, wwr)
        windows = idf.idfobjects['FENESTRATIONSURFACE:DETAILED']
        assert len(windows) == 8
        for window in windows:
            wall = idf.getobject('BUILDINGSURFACE:DETAILED',
                                 window.Building_Surface_Name)
            assert almostequal(window.area, wall.area * wwr, 3)


class TestViewGeometry():

    def test_get_surfaces(self, base_idf):
        # type: () -> None
        idf = base_idf
        surfaces = _get_surfaces(idf)
        assert len(surfaces) == 12

    def test_get_shading(self, base_idf):
        # type: () -> None
        idf = base_idf
        shading = _get_shading(idf)
        assert len(shading) == 1

    def test_get_collections(self, base_idf):
        # type: () -> None
        idf = base_idf
        collections = _get_collections(idf)
        assert len(collections) == 5
        for c in collections:
            assert c

    def test_view_model(self, base_idf):
        # type: () -> None
        idf = base_idf
        idf.view_model(test=True)


@pytest.fixture()
def new_idf():
    IDF.iddname == None
    iddfhandle = StringIO(iddcurrent.iddtxt)
    IDF.setiddname(iddfhandle, testing=True)
    idf = IDF()
    idf.new()
    return idf

@pytest.fixture()
def wwr_idf(new_idf):
    test_walls = [[0,0,0], [0,1,0], [0,1,1], [0,0,1]], [[0,0,0], [1,0,0], [1,0,1], [0,0,1]]
    for i, coords in enumerate(test_walls, 1):
        new_idf.newidfobject(
            'FENESTRATIONSURFACE:DETAILED',
            Name='window%s' % i,
            Surface_Type='window',
            Construction_Name='ExtWindow',
            Building_Surface_Name='wall%s' % i,
        )
        wall = new_idf.newidfobject(
            'BUILDINGSURFACE:DETAILED',
            Name='wall%s' % i,
            Surface_Type='wall',
            Outside_Boundary_Condition='Outdoors',
        )
        wall.setcoords(coords)
    return new_idf


class TestWWR:

    def is_expected_wwr(self, idf, wwr):
        windows_area = sum(w.area for w in idf.getsubsurfaces('window'))
        walls_area = sum(w.area for w in idf.getsurfaces('wall'))
        return almostequal(windows_area, walls_area * wwr, 3)

    def test_wwr(self, wwr_idf):
        idf = wwr_idf
        wwr = 0.2
        idf.set_wwr(wwr)
        assert self.is_expected_wwr(idf, wwr), 'Not all walls have the expected WWR set'

    def test_wwr_map(self, wwr_idf):
        idf = wwr_idf
        expected_wwr = 0.2
        idf.set_wwr(wwr=0.99, wwr_map={90.0: 0.1, 180: 0.3})
        assert self.is_expected_wwr(idf, expected_wwr), 'Not all walls have the expected WWR set'

    def test_wwr_zero(self, wwr_idf):
        idf = wwr_idf
        idf.set_wwr(wwr=0, wwr_map={90.0: 0.1})
        expected_wwr = 0.05
        assert self.is_expected_wwr(idf, expected_wwr), 'Not all walls have the expected WWR set'

    def test_wwr_none(self, wwr_idf):
        idf = wwr_idf
        idf.set_wwr(wwr=None, wwr_map={90.0: 0.3})
        expected_wwr = 0.15
        assert self.is_expected_wwr(idf, expected_wwr), 'Not all walls have the expected WWR set'

    def test_wwr_two_window_constructions(self, wwr_idf):
        idf = wwr_idf
        idf.newidfobject(
            'FENESTRATIONSURFACE:DETAILED',
            Name='window2',
            Surface_Type='window',
            Construction_Name='ExtWindow2',
            Building_Surface_Name='wall1',
        )
        wwr = 0.2
        try:
            idf.set_wwr(wwr)
            assert False, 'Should have raised an error since windows with more than one construction are present'
        except ValueError:
            pass
        idf.set_wwr(wwr, construction='ExtWindow')
        assert self.is_expected_wwr(idf, wwr)

    def test_wwr_mixed_subsurfaces(self, wwr_idf):
        idf = wwr_idf
        idf.newidfobject(
            'DOOR',
            Name='door1',
            Construction_Name='ExtDoor',
            Building_Surface_Name='wall1',
        )
        wwr = 0.2
        try:
            idf.set_wwr(wwr)
            assert False, 'Should have raised an error since not all subsurfaces are windows'
        except ValueError:
            pass
        idf.set_wwr(wwr, force=True)
        assert self.is_expected_wwr(idf, wwr)



# Copyright (c) 2016 Jamie Bull
# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""
Monkey patches for changes to classes and functions in Eppy. These include
fixes which have not yet made it to the released version of Eppy. These will be
removed if/when they are added to Eppy.

"""
import copy
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union  # noqa

from eppy import bunchhelpers, iddgaps
from eppy.EPlusInterfaceFunctions import eplusdata, iddindex, parse_idd
from eppy.EPlusInterfaceFunctions.eplusdata import Eplusdata  # noqa
from eppy.bunch_subclass import EpBunch as BaseBunch
from eppy.idf_msequence import Idf_MSequence
from eppy.idfreader import convertallfields, iddversiontuple
from eppy.modeleditor import IDF as BaseIDF
from eppy.modeleditor import IDDNotSetError, namebunch, newrawobject

from .geom.polygons import Polygon3D  # noqa
from .geom.surfaces import set_coords
from .geom.vectors import Vector3D  # noqa
if False: from .idf import IDF  # noqa


class EpBunch(BaseBunch):
    """Monkeypatched EpBunch to add the setcoords function."""

    def setcoords(self,
                  poly,  # type: Union[List[Vector3D], List[Tuple[float, float, float]], Polygon3D]
                  ggr=None  # type: Union[List, None, Idf_MSequence]
                  ):
        # type: (...) -> None
        """Set the coordinates of a surface.

        :param poly: Either a Polygon3D object of a list of (x,y,z) tuples.
        :param ggr: A GlobalGeometryRules IDF object. Defaults to None.

        """
        surfaces = [
            "BUILDINGSURFACE:DETAILED",
            "WALL:DETAILED",
            "ROOFCEILING:DETAILED",
            "FLOOR:DETAILED",
            "FENESTRATIONSURFACE:DETAILED",
            "SHADING:SITE:DETAILED",
            "SHADING:BUILDING:DETAILED",
            "SHADING:ZONE:DETAILED", ]
        if self.key.upper() in surfaces:
            set_coords(self, poly, ggr)
        else:
            raise AttributeError


def idfreader1(fname,  # type: str
               iddfile,  # type: str
               theidf,  # type: IDF
               conv=True,  # type: Optional[bool]
               commdct=None,  # type: Optional[List[List[Dict[str, Any]]]]
               block=None  # type: Optional[List]
               ):
    # type: (...) -> Tuple[Dict[str, Idf_MSequence], List, Eplusdata, List[List[Dict[str, Any]]], Dict, Tuple[int]]
    """Read idf file and return bunches.

    :param fname: Name of the IDF file to read.
    :param iddfile: Name of the IDD file to use to interpret the IDF.
    :param conv: If True, convert strings to floats and integers where marked in the IDD. Defaults to None.
    :param commdct: Descriptions of IDF fields from the IDD. Defaults to None.
    :param block: EnergyPlus field ID names of the IDF from the IDD. Defaults to None.
    :returns: bunchdt Dict of lists of idf_MSequence objects in the IDF.
    :returns: block EnergyPlus field ID names of the IDF from the IDD.
    :returns data: Eplusdata object containing representions of IDF objects.
    :returns: commdct List of names of IDF objects.
    :returns: idd_index A pair of dicts used for fast lookups of names of groups of objects.
    :returns: versiontuple Version of EnergyPlus from the IDD.

    """
    versiontuple = iddversiontuple(iddfile)
    block, data, commdct, idd_index = readdatacommdct1(
        fname,
        iddfile=iddfile,
        commdct=commdct,
        block=block)
    if conv:
        convertallfields(data, commdct)
    # fill gaps in idd
    if versiontuple < (8,):
        skiplist = ["TABLE:MULTIVARIABLELOOKUP"]
    else:
        skiplist = None
    nofirstfields = iddgaps.missingkeys_standard(
        commdct, data.dtls,
        skiplist=skiplist)
    iddgaps.missingkeys_nonstandard(block, commdct, data.dtls, nofirstfields)
    bunchdt = makebunches(data, commdct, theidf)

    return bunchdt, block, data, commdct, idd_index, versiontuple


def readdatacommdct1(idfname,   # type: str
                     iddfile='Energy+.idd',   # type: str
                     commdct=None,  # type: Optional[List[List[Dict[str, Any]]]]
                     block=None  # type: Optional[List]
                     ):
    # type: (...) -> Tuple[List, Eplusdata, List[List[Dict[str, Any]]], Dict]
    """Read the idf file.

    This is patched so that the IDD index is not lost when reading a new IDF without reloading the modeleditor module.

    :param idfname: Name of the IDF file to read.
    :param iddfile: Name of the IDD file to use to interpret the IDF.
    :param commdct: Descriptions of IDF fields from the IDD. Defaults to None.
    :param block: EnergyPlus field ID names of the IDF from the IDD. Defaults to None.
    :returns: block EnergyPlus field ID names of the IDF from the IDD.
    :returns data: Eplusdata object containing representions of IDF objects.
    :returns: commdct List of names of IDF objects.
    :returns: idd_index A pair of dicts used for fast lookups of names of groups of objects.

    """
    if not commdct:
        block, commlst, commdct, idd_index = parse_idd.extractidddata(iddfile)
        theidd = eplusdata.Idd(block, 2)
    else:
        theidd = eplusdata.Idd(block, 2)
        name2refs = iddindex.makename2refdct(commdct)
        ref2namesdct = iddindex.makeref2namesdct(name2refs)
        idd_index = dict(name2refs=name2refs, ref2names=ref2namesdct)
        commdct = iddindex.ref2names2commdct(ref2namesdct, commdct)
    data = eplusdata.Eplusdata(theidd, idfname)
    return block, data, commdct, idd_index


def addthisbunch(bunchdt,  # type: Dict[str, Idf_MSequence]
                 data,  # type: Eplusdata
                 commdct,  # type: Optional[List[List[Dict[str, Any]]]]
                 thisbunch,  # type: EpBunch
                 _idf  # type: IDF
                 ):
    # type: (...) -> EpBunch
    """Add an object to the IDF. Monkeypatched to return the object.

    `thisbunch` usually comes from another idf file or it can be used to copy within the idf file.

    :param bunchdt: Dict of lists of idf_MSequence objects in the IDF.
    :param data: Eplusdata object containing representions of IDF objects.
    :param commdct: Descriptions of IDF fields from the IDD.
    :param thisbunch: The object to add to the model.
    :param _idf: The IDF object. Not used either here or in Eppy but kept for consistency with Eppy.
    :returns: The EpBunch object added.

    """
    key = thisbunch.key.upper()
    obj = copy.copy(thisbunch.obj)
    abunch = obj2bunch(data, commdct, obj)
    bunchdt[key].append(abunch)
    return abunch


def makebunches(data,  # type: Eplusdata
                commdct,  # type: Optional[List[List[Dict[str, Any]]]]
                theidf  # type: IDF
                ):
    # type: (...) -> Dict[str, Idf_MSequence]
    """Make bunches with data.

    :param data: Eplusdata object containing representions of IDF objects.
    :param commdct: Descriptions of IDF fields from the IDD.
    :param theidf: The IDF object.
    :returns: Dict of lists of idf_MSequence objects in the IDF.

    """
    bunchdt = {}
    dt, dtls = data.dt, data.dtls
    for obj_i, key in enumerate(dtls):
        key = key.upper()
        objs = dt[key]
        list1 = []
        for obj in objs:
            bobj = makeabunch(commdct, obj, obj_i)
            list1.append(bobj)
        bunchdt[key] = Idf_MSequence(list1, objs, theidf)
    return bunchdt


def obj2bunch(data,  # type: Eplusdata
              commdct,  # type: Optional[List[List[Dict[str, Any]]]]
              obj  # type: List[str]
              ):
    # type: (...) -> EpBunch
    """Make a new bunch object using the data object.

    :param data: Eplusdata object containing representions of IDF objects.
    :param commdct: Descriptions of IDF fields from the IDD.
    :param obj: List of field values in an object.
    :returns: EpBunch object.

    """
    dtls = data.dtls
    key = obj[0].upper()
    key_i = dtls.index(key)
    abunch = makeabunch(commdct, obj, key_i)
    return abunch


def makeabunch(commdct,  # type: List[List[Dict[str, Any]]]
               obj,  # type: Union[List[Union[float, str]], List[str]]
               obj_i  # type: int
               ):
    # type: (...) -> EpBunch
    """Make a bunch from the object.

    :param commdct: Descriptions of IDF fields from the IDD.
    :param obj: List of field values in an object.
    :param obj_i: Index of the object in commdct.
    :returns: EpBunch object.

    """
    objidd = commdct[obj_i]
    objfields = [comm.get('field') for comm in commdct[obj_i]]
    objfields[0] = ['key']
    objfields = [field[0] for field in objfields]
    obj_fields = [bunchhelpers.makefieldname(field) for field in objfields]
    bobj = EpBunch(obj, obj_fields, objidd)
    return bobj


class PatchedIDF(BaseIDF):
    """Monkey-patched IDF.

    Patched to add read (to add additional functionality) and to fix copyidfobject and newidfobject.
    """

    def read(self):
        """Read the IDF file and the IDD file.

        If the IDD file had already been read, it will not be read again.

        Populates the following data structures::

        - idfobjects : list
        - model : list
        - idd_info : list
        - idd_index : dict
        """
        if self.getiddname() is None:
            errortxt = (
                "IDD file needed to read the idf file. Set it using IDF.setiddname(iddfile)")
            raise IDDNotSetError(errortxt)
        self.idfobjects, block, self.model, idd_info, idd_index, versiontuple = idfreader1(
            self.idfname,
            self.iddname,
            self,
            commdct=self.idd_info,
            block=self.block,
        )
        self.__class__.setidd(idd_info, idd_index, block, versiontuple)

    def newidfobject(self, key, aname='', **kwargs):
        # type: (str, str, **Any) -> EpBunch
        """Add a new idfobject to the model.

        If you don't specify a value for a field, the default value will be set.

        For example ::

            newidfobject("CONSTRUCTION")
            newidfobject("CONSTRUCTION",
                Name='Interior Ceiling_class',
                Outside_Layer='LW Concrete',
                Layer_2='soundmat')

        :param key: The type of IDF object. This must be in ALL_CAPS.
        :param aname: This parameter is not used. It is left there for backward compatibility.
        :param kwargs: Keyword arguments in the format `field=value` used to set fields in the EnergyPlus object.
        :returns: EpBunch object.
        """
        obj = newrawobject(self.model, self.idd_info, key)
        abunch = obj2bunch(self.model, self.idd_info, obj)
        if aname:
            warnings.warn(
                "The aname parameter should no longer be used (%s)." % aname, UserWarning)
            namebunch(abunch, aname)
        self.idfobjects[key].append(abunch)  # type: Dict[str, Idf_MSequence]
        for k, v in kwargs.items():
            abunch[k] = v
        return abunch

    def copyidfobject(self, idfobject):
        # type: (EpBunch) -> EpBunch
        """Add an IDF object to the IDF.

        This has been monkey-patched to add the return value.

        :param idfobject: The IDF object to copy. Usually from another IDF, or it can be used to copy within this IDF.
        :returns: EpBunch object.
        """
        return addthisbunch(self.idfobjects, self.model, self.idd_info, idfobject, self)



"""
This module contains the implementation of `geomeppy.IDF`.
"""
from typing import Any, Dict, List, Optional, Union  # noqa

from eppy.bunch_subclass import EpBunch  # noqa
from eppy.idf_msequence import Idf_MSequence  # noqa

from .geom.intersect_match import intersect_idf_surfaces, match_idf_surfaces
from .builder import Block, Zone
from .geom.polygons import bounding_box, Polygon2D  # noqa
from .geom.vectors import Vector2D, Vector3D  # noqa
from .io.obj import export_to_obj
from .patches import PatchedIDF
from .recipes import set_default_constructions, set_wwr, rotate, scale, translate, translate_to_origin
from .view_geometry import view_idf


def new_idf(fname):
    # type (Optional[str]) -> IDF
    """Create a new blank IDF.

    :param fname: A name for the new IDF.

    """
    idf = IDF()
    idf.new(fname)
    return idf


class IDF(PatchedIDF):
    """Geometry-enabled IDF class, usable in the same way as Eppy's IDF.

    This adds geometry functionality to Eppy's IDF class.

    """

    def intersect_match(self):
        # type: () -> None
        """Intersect all surfaces in the IDF, then set boundary conditions."""
        self.intersect()
        self.match()

    def intersect(self):
        # type: () -> None
        """Intersect all surfaces in the IDF."""
        intersect_idf_surfaces(self)

    def match(self):
        # type: () -> None
        """Set boundary conditions for all surfaces in the IDF."""
        match_idf_surfaces(self)

    def translate_to_origin(self):
        # type: () -> None
        """Move an IDF close to the origin so that it can be viewed in SketchUp."""
        translate_to_origin(self)

    def translate(self, vector):
        # type: (Vector2D) -> None
        """Move the IDF in the direction given by a vector.

        :param vector: A vector to translate by.

        """
        surfaces = self.getsurfaces()
        translate(surfaces, vector)
        subsurfaces = self.getsubsurfaces()
        translate(subsurfaces, vector)
        shadingsurfaces = self.getshadingsurfaces()
        translate(shadingsurfaces, vector)

    def rotate(self, angle, anchor=None):
        # type: (Union[int, float], Optional[Union[Vector2D, Vector3D]]) -> None
        """Rotate the IDF counterclockwise by the angle given.

        :param angle: Angle (in degrees) to rotate by.
        :param anchor: Point around which to rotate. Default is the centre of the the IDF's bounding box.

        """
        anchor = anchor or self.centroid
        surfaces = self.getsurfaces()
        subsurfaces = self.getsubsurfaces()
        shadingsurfaces = self.getshadingsurfaces()
        self.translate(-anchor)
        rotate(surfaces, angle)
        rotate(subsurfaces, angle)
        rotate(shadingsurfaces, angle)
        self.translate(anchor)

    def scale(self, factor, anchor=None):
        # type: (Union[int, float], Optional[Union[Vector2D, Vector3D]]) -> None
        """Scale the IDF by a scaling factor.

        :param factor: Factor to scale by.
        :param anchor: Point to scale around. Default is the centre of the the IDF's bounding box.

        """
        anchor = anchor or self.centroid
        surfaces = self.getsurfaces()
        subsurfaces = self.getsubsurfaces()
        shadingsurfaces = self.getshadingsurfaces()
        self.translate(-anchor)
        scale(surfaces, factor)
        scale(subsurfaces, factor)
        scale(shadingsurfaces, factor)
        self.translate(anchor)

    def set_default_constructions(self):
        # type: () -> None
        set_default_constructions(self)

    def getsurfaces(self, surface_type=None):
        # type: (Optional[str]) -> Union[List[EpBunch], Idf_MSequence]
        """Return all surfaces in the IDF.

        :param surface: Type of surface to get. Defaults to all.
        :returns: IDF surfaces.

        """
        surfaces = self.idfobjects['BUILDINGSURFACE:DETAILED']
        if surface_type:
            surfaces = [s for s in surfaces if s.Surface_Type.lower() ==
                        surface_type.lower()]
        return surfaces

    def bounding_box(self):
        # type: () -> Polygon2D
        """Calculate the site bounding box.

        :returns: A polygon of the bounding box.

        """
        floors = self.getsurfaces('floor')
        return bounding_box(floors)

    @property
    def centroid(self):
        # type: () -> Vector2D
        """Calculate the centroid of the site bounding box.

        :returns: The centroid of the site bounding box.

        """
        bbox = self.bounding_box()
        return bbox.centroid

    def getsubsurfaces(self, surface_type=None):
        # type: (Optional[str]) -> Union[List[EpBunch], Idf_MSequence]
        """Return all subsurfaces in the IDF.

        :returns: IDF surfaces.

        """
        surfaces = self.idfobjects['FENESTRATIONSURFACE:DETAILED']
        if surface_type:
            surfaces = [s for s in surfaces
                        if s.Surface_Type.lower() == surface_type.lower()]
        return surfaces

    def getshadingsurfaces(self, surface_type=None):
        # type: (Optional[str]) -> Union[List[EpBunch], Idf_MSequence]
        """Return all subsurfaces in the IDF.

        :returns: IDF surfaces.

        """
        surfaces = self.idfobjects['SHADING:ZONE:DETAILED']
        if surface_type:
            surfaces = [s for s in surfaces
                        if s.Surface_Type.lower() == surface_type.lower()]
        return surfaces

    def set_wwr(self, wwr=0.2, construction=None, force=False, wwr_map={}):
        # type: (Optional[float], Optional[str], Optional[bool], Optional[dict]) -> None
        """Add strip windows to all external walls.

        Different WWR can be applied to specific wall orientations using the `wwr_map` keyword arg.
        This map is a dict of wwr values, keyed by `wall.azimuth`, which overrides the default passed as `wwr`.

        :param wwr: Window to wall ratio in the range 0.0 to 1.0.
        :param construction: Name of a window construction.
        :param force: True to remove all subsurfaces before setting the WWR.
        :param wwr_map: Mapping from wall orientation (azimuth) to WWR, e.g. {180: 0.25, 90: 0.2}.

        """
        set_wwr(self, wwr, construction, force, wwr_map)

    def view_model(self, test=False):
        # type: (Optional[bool]) -> None
        """Show a zoomable, rotatable representation of the IDF."""
        view_idf(idf_txt=self.idfstr(), test=test)

    def to_obj(self, fname=None, mtllib=None):
        # type: (Optional[str], Optional[str]) -> None
        """Export an OBJ file representation of the IDF.

        This can be used for viewing in tools which support the .obj format.

        :param fname: A filename for the .obj file. If None we try to base it on IDF.idfname and change the filetype.
        :param mtllib: The name of a .mtl file to be referenced from the .obj file. If None, we use default.mtl.
        """
        if not fname:
            try:
                fname = self.idfname.replace('.idf', '.obj')
            except AttributeError:
                fname = 'default.obj'
        export_to_obj(self, fname, mtllib)

    def add_block(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        """Add a block to the IDF.

        :param name: A name for the block.
        :param coordinates: A list of (x, y) tuples representing the building outline.
        :param height: The height of the block roof above ground level.
        :param num_stories: The total number of stories including basement stories. Default : 1.
        :param below_ground_stories: The number of stories below ground. Default : 0.
        :param below_ground_storey_height: The height of each basement storey. Default : 2.5.

        """
        block = Block(*args, **kwargs)
        zoning = kwargs.get('zoning', 'by_storey')
        if zoning == 'by_storey':
            zones = [Zone('Block %s Storey %i' %
                          (block.name, storey['storey_no']), storey)
                     for storey in block.stories]
        else:
            raise ValueError('%s is not a valid zoning rule' % zoning)
        for zone in zones:
            self.add_zone(zone)

    def add_shading_block(self, *args, **kwargs):
        # type: (*Any, **Any) -> None
        """Add a shading block to the IDF.

        :param name: A name for the block.
        :param coordinates: A list of (x, y) tuples representing the building outline.
        :param height: The height of the block roof above ground level.
        :param num_stories: The total number of stories including basement stories. Default : 1.
        :param below_ground_stories: The number of stories below ground. Default : 0.
        :param below_ground_storey_height: The height of each basement storey. Default : 2.5.

        """
        block = Block(*args, **kwargs)
        for i, wall in enumerate(block.walls[0], 1):
            if wall.area <= 0:
                continue
            s = self.newidfobject(
                'SHADING:SITE:DETAILED',
                Name='%s_%s' % (block.name, i),
            )
            try:
                s.setcoords(wall)
            except ZeroDivisionError:
                self.removeidfobject(s)

    def add_zone(self, zone):
        # type: (Zone) -> None
        """Add a zone to the IDF.

        :param zone: A Zone object holding details about the zone.

        """
        try:
            ggr = self.idfobjects['GLOBALGEOMETRYRULES'][0]  # type: Dict[str, Idf_MSequence]
        except IndexError:
            ggr = None
        # add zone object
        self.newidfobject('ZONE', Name=zone.name)

        for surface_type in zone.__dict__.keys():
            if surface_type == 'name':
                continue
            for i, surface_coords in enumerate(zone.__dict__[surface_type], 1):
                if not surface_coords:
                    continue
                name = '{name} {s_type} {num:04d}'.format(name=zone.name, s_type=surface_type[:-1].title(), num=i)
                s = self.newidfobject(
                    'BUILDINGSURFACE:DETAILED',
                    Name=name,
                    Surface_Type=surface_type[:-1],
                    Zone_Name=zone.name,
                )
                s.setcoords(surface_coords, ggr)



"""2D and 3D vector classes.

These are used to represent points in 2D and 3D, as well as directions for translations.
"""

from typing import Any, Iterable, Iterator, List, Sized, Tuple, Union  # noqa

import numpy as np
from six.moves import zip

if False: from .polygons import Polygon3D  # noqa


class Vector2D(Sized, Iterable):
    """Two dimensional point."""

    def __init__(self, *args):
        # type: (*Any) -> None
        self.args = list(args)
        self.x = float(args[0])
        self.y = float(args[1])
        self.z = 0.0

    def __iter__(self):
        # type: () -> Iterator
        return (i for i in self.args)

    def __repr__(self):
        # type: () -> str
        class_name = type(self).__name__
        return '{}({!r}, {!r})'.format(class_name, *self.args)

    def __eq__(self, other):
        for a, b in zip(self, other):
            if a != b:
                return False
        return True

    def __sub__(self, other):
        # type: (Any) -> Union[Vector2D, Vector3D]
        return self.__class__(*[self[i] - other[i] for i in range(len(self))])

    def __add__(self, other):
        # type: (Any) -> Union[Vector2D, Vector3D]
        return self.__class__(*[self[i] + other[i] for i in range(len(self))])

    def __neg__(self):
        # type: () -> Union[Vector2D, Vector3D]
        return self.__class__(*inverse_vector(self))

    def __len__(self):
        # type: () -> int
        return len(self.args)

    def __getitem__(self, key):
        # type: (Union[int, slice]) -> Union[Tuple[float, float, float], float]
        return self.args[key]

    def __setitem__(self, key, value):
        self.args[key] = value

    def __hash__(self):
        return hash(self.x) ^ hash(self.y)

    def dot(self, other):
        # type: (Vector3D) -> np.float64
        return np.dot(self, other)

    def cross(self, other):
        # type: (Union[Vector2D, Vector3D]) -> np.ndarray
        return np.cross(self, other)

    @property
    def length(self):
        # type: () -> float
        """The length of a vector."""
        length = sum(x ** 2 for x in self.args) ** 0.5

        return length

    def closest(self, poly):
        # type: (Polygon3D) -> Union[Vector2D, Vector3D]
        """Find the closest vector in a polygon.

        :param poly: Polygon or Polygon3D

        """
        min_d = float('inf')
        closest_pt = None
        for pt2 in poly:
            direction = self - pt2
            sq_d = sum(x ** 2 for x in direction)
            if sq_d < min_d:
                min_d = sq_d
                closest_pt = pt2
        return closest_pt

    def normalize(self):
        # type: () -> Union[Vector2D, Vector3D]
        return self.set_length(1.0)

    def set_length(self, new_length):
        # type: (float) -> Union[Vector2D, Vector3D]
        current_length = self.length
        multiplier = new_length / current_length
        self.args = [i * multiplier for i in self.args]
        return self

    def invert(self):
        # type: () -> Union[Vector2D, Vector3D]
        return -self

    def as_array(self, dims=3):
        # type: (Union[Vector2D, Vector3D], int) -> np.ndarray
        """Convert a point to a numpy array.

        Converts a Vector3D to a numpy.array([x,y,z]) or a Vector2D to a numpy.array([x,y]).
        Ensures all values are floats since some other types cause problems in pyclipper (notably where sympy.Zero is
        used to represent 0.0).

        :param pt: The point to convert.
        :param dims: Number of dimensions {default : 3}.
        :returns: Vector as a Numpy array.

        """
        # handle Vector3D
        if dims == 3:
            return np.array([float(self.x), float(self.y), float(self.z)])
        # handle Vector2D
        elif dims == 2:
            return np.array([float(self.x), float(self.y)])
        else:
            raise ValueError('%s-dimensional vectors are not supported.' % dims)

    def as_tuple(self, dims=3):
        # type: (Union[Vector2D, Vector3D, int]) -> Union[Tuple[float, float], Tuple[float, float, float]]
        """Convert a point to a numpy array.

        Convert a Vector3D to an (x,y,z) tuple or a Vector2D to an (x,y) tuple.
        Ensures all values are floats since some other types cause problems in pyclipper (notably where sympy.Zero is
        used to represent 0.0).

        :param pt: The point to convert.
        :param dims: Number of dimensions {default : 3}.
        :returns: Vector as a tuple.

        """
        # handle Vector3D
        if dims == 3:
            return float(self.x), float(self.y), float(self.z)
        # handle Vector2D
        elif dims == 2:
            return float(self.x), float(self.y)
        else:
            raise ValueError('%s-dimensional vectors are not supported.' % dims)

    def relative_distance(self, v2):
        # type: (Vector3D) -> float
        """A distance function for sorting vectors by distance.

        This only provides relative distance, not actual distance since we only use it for sorting.

        :param v2: Another vector.
        :returns: Relative distance between two point vectors.

        """
        direction = self - v2
        return sum(x ** 2 for x in direction)


class Vector3D(Vector2D):
    """Three dimensional point."""

    def __init__(self,
                 x,  # type: Union[float, np.float64]
                 y,  # type: Union[float, np.float64]
                 z=0,  # type: Union[float, np.float64]
                 ):
        # type: (...) -> None
        super(Vector3D, self).__init__(x, y, z)
        self.z = float(z)
        self.args = [self.x, self.y, self.z]

    def __repr__(self):
        # type: () -> str
        class_name = type(self).__name__
        return '{}({!r}, {!r}, {!r})'.format(class_name, *self.args)

    def __hash__(self):
        # type: () -> int
        return hash(self.x) ^ hash(self.y) ^ hash(self.z)


def inverse_vector(v):
    # type: (Union[Vector2D, Vector3D]) -> List[float]
    """Convert a vector to the same vector but in the opposite direction

    :param v: The vector.
    :returns: The vector reversed.

    """
    return [-i for i in v]



from eppy.iddcurrent import iddcurrent
import pytest
from six import StringIO

from geomeppy.idf import IDF
from geomeppy.io.obj import export_to_obj

idf_txt = """
Version, 8.5;
Building, Building 1, , , , , , , ;
Zone, z1 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
Zone, z2 Thermal Zone, 0.0, 0.0, 0.0, 0.0, , 1, , , , , , Yes;
BuildingSurface:Detailed, z1_FLOOR, Floor, , z1 Thermal Zone, ground, , NoSun, NoWind, , , 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0;
BuildingSurface:Detailed, z1_ROOF, Roof, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 2.0, 0.5, 1.0, 2.1, 0.5, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0001, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.0, 1.1, 0.5, 1.0, 1.1, 0.0, 1.0, 2.1, 0.0, 1.0, 2.1, 0.5;
BuildingSurface:Detailed, z1_WALL_0002, Wall, , z1 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 1.0, 2.1, 0.5, 1.0, 2.1, 0.0, 2.0, 2.0, 0.0, 2.0, 2.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0003, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 2.0, 0.5, 2.0, 2.0, 0.0, 2.0, 1.0, 0.0, 2.0, 1.0, 0.5;
BuildingSurface:Detailed, z1_WALL_0004, WALL, , z1 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.0, 1.0, 0.5, 2.0, 1.0, 0.0, 1.0, 1.1, 0.0, 1.0, 1.1, 0.5;
BuildingSurface:Detailed, z2_FLOOR, Floor, , z2 Thermal Zone, ground, , NoSun, NoWind, , , 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0;
BuildingSurface:Detailed, z2_ROOF, Roof, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 2.95, 0.5, 1.5, 3.05, 0.5, 1.5, 2.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0001, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 2.05, 0.5, 1.5, 2.05, 0.0, 1.5, 3.05, 0.0, 1.5, 3.05, 0.5;
BuildingSurface:Detailed, z2_WALL_0002, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 1.5, 3.05, 0.5, 1.5, 3.05, 0.0, 2.5, 2.95, 0.0, 2.5, 2.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0003, WALL, , z2 Thermal Zone, outdoors, , SunExposed, WindExposed, , , 2.5, 2.95, 0.5, 2.5, 2.95, 0.0, 2.5, 1.95, 0.0, 2.5, 1.95, 0.5;
BuildingSurface:Detailed, z2_WALL_0004, Wall, , z2 Thermal Zone, Outdoors, , SunExposed, WindExposed, , , 2.5, 1.95, 0.5, 2.5, 1.95, 0.0, 1.5, 2.05, 0.0, 1.5, 2.05, 0.5;
Shading:Zone:Detailed, z2_SHADE_0003, z2_WALL_0003, , 4, 2.5, 2.95, 0.5, 2.6, 2.95, 0.3, 2.6, 1.95, 0.3, 2.5, 1.95, 0.5;
"""


@pytest.fixture()
def base_idf():
    # type: () -> None
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() == None:
        IDF.setiddname(iddfhandle)
    idf = IDF(StringIO(idf_txt))
    idf.intersect_match()
    idf.set_wwr(0.25)
    return idf


def test_export_to_obj(base_idf):
    export_to_obj(base_idf, 'test.obj')



"""Module for extracting the geometry from an existing IDF.

There is the option to copy:

- thermal zone description and geometry
- surface construction elements

"""
import itertools
from typing import Optional  # noqa

from .idf import EpBunch, Idf_MSequence, IDF, new_idf  # noqa


def copy_constructions(source_idf, target_idf=None, fname=None):
    # type: (IDF, Optional[IDF], str) -> IDF
    """Extract construction objects from a source IDF and add them to a target IDF or a new IDF.

    :param source_idf: An IDF to source objects from.
    :param target_idf: An optional IDF to add objects to. If none is passed, a new IDF is returned.
    :param fname: A name for the new IDF created if no target IDF is passed in.
    :returns: Either the target IDF or a new IDF containing the construction objects.
    """
    group = 'Surface Construction Elements'
    if not target_idf:
        target_idf = new_idf(fname)
    idf = copy_group(source_idf, target_idf, group)
    return idf


def copy_geometry(source_idf, target_idf=None, fname=None):
    # type: (IDF, Optional[IDF], str) -> IDF
    """Extract geometry objects from a source IDF and add them to a target IDF or a new IDF..

    :param source_idf: An IDF to source objects from.
    :param target_idf: An optional IDF to add objects to. If none is passed, a new IDF is returned.
    :param fname: A name for the new IDF created if no target IDF is passed in.
    :returns: Either the target IDF or a new IDF containing the geometry objects.
    """
    group = 'Thermal Zones and Surfaces'
    if not target_idf:
        target_idf = new_idf(fname)
    idf = copy_group(source_idf, target_idf, group)
    return idf


def copy_group(source_idf, target_idf, group):
    # type: (IDF, IDF, str) -> IDF
    """Extract a group of objects from a source IDF and add them to a target IDF.

    :param source_idf: An IDF to source objects from.
    :param target_idf: An IDF to add objects to.
    :param group: The name of the group of objects to copy.
    :returns: A new IDF containing the objects which belong to the group.
    """
    keys = source_idf.getiddgroupdict()[group]
    objects = [source_idf.idfobjects[key.upper()] for key in keys]
    for obj in itertools.chain(*objects):
        target_idf.copyidfobject(obj)
    return target_idf



"""
Module for export of .obj files
-------------------------------

The OBJ file format developed by Wavefront Technologies is used in programs like Blender to represent 3D models.

`Geomeppy` is able to output simple .obj files, and accompanying .mtl files to represent materials. This module
contains the source code for this operation. It should be treated as developers' reference only. As a user, if you
want to create a .obj file it is best to use the `IDF.to_obj()` function.

The generated files can be viewed online at https://3dviewer.net/. Drag the .obj file and the .mtl file into the
browser and you will be able to interact with a zoomable, rotatable model of your IDF.

`OBJ file specifications <https://www.cs.utah.edu/~boulos/cs3505/obj_spec.pdf>`_

`MTL file specifications <http://people.sc.fsu.edu/~jburkardt/data/mtl/mtl.html>`_

Example polygon::

    # vertices
    v 0.0 0.0 0.0
    v 1.0 0.0 0.0
    v 1.0 0.0 1.0
    v 0.0 0.0 1.0

    # face
    f 1// 2// 3// 4//

"""
from itertools import product
import shutil
from typing import List, Optional, Set  # noqa
import os

import pypoly2tri as p2t

if False: from ..idf import IDF  # noqa
from ..geom.polygons import Polygon2D, Polygon3D
from ..geom.vectors import Vector3D  # noqa

THIS_DIR = os.path.abspath(os.path.dirname(__file__))


class ObjWriter(object):
    """Container class holding the data needed to generate the .obj file."""
    faces = []  # type: List[int]
    vertices = []  # type: List[Vector3D]
    # v_set is used for fast testing if a vertex is in vertices
    v_set = set()  # type: Set[Vector3D]

    def from_surfaces(self, surfaces, subsurfaces, shading_surfaces):
        self.prepare_surfaces(surfaces, subsurfaces)
        self.prepare_shadingsurfaces(shading_surfaces)

    def prepare_surfaces(self, surfaces, subsurfaces):
        for i, surface in enumerate(surfaces):
            face_subsurfaces = [ss for ss in subsurfaces if ss.Building_Surface_Name == surface.Name]
            if face_subsurfaces:
                subsurface = face_subsurfaces[0]
                self.build_surface_with_subsurface(surface, subsurface)
            else:
                self.build_simple_surface(surface)

    def build_simple_surface(self, surface):
        poly = Polygon3D(surface.coords)
        poly2d = poly.project_to_2D()
        if poly2d.is_convex:
            # no need to triangulate the surface
            self.add_face(surface.coords, surface.Surface_Type)
            return
        coords = [p2t.shapes.Point(x, y) for x, y in poly2d.vertices]
        cdt = p2t.cdt.CDT(coords)
        cdt.Triangulate()
        triangles = cdt.GetTriangles()
        for t in triangles:
            tri2d = Polygon2D([(p.x, p.y) for p in t.points_])
            tri3d = tri2d.project_to_3D(poly)
            self.add_face(tri3d, surface.Surface_Type)
        return

    def build_surface_with_subsurface(self, surface, subsurface):
        outer_poly = Polygon3D(surface.coords)
        inner_poly = Polygon3D(subsurface.coords)
        for edge in outer_poly.edges:
            links = product(edge, inner_poly)
            links = sorted(links, key=lambda x: x[0].relative_distance(x[1]))
            pt1, pt2 = edge
            t1 = [pt1, links[0][1], pt2]
            t2 = [pt2, t1[1], links[1][1]]
            self.add_face(t1, surface.Surface_Type)
            self.add_face(t2, surface.Surface_Type)
        self.add_face(subsurface.coords, subsurface.Surface_Type, test=False)

    def prepare_shadingsurfaces(self, shading_surfaces):
        for s in shading_surfaces:
            self.add_face(s.coords, 'shade')

    def add_face(self, coords, mtl, test=True):
        face = []
        for v in coords:
            if test and v not in self.v_set:
                self.vertices.append(v)
                self.v_set.add(v)
                face.append(len(self.vertices))
            else:
                face.append(self.vertices.index(v) + 1)
        self.faces.append({'face': face, 'mtl': mtl})

    def write(self, fname, mtllib):
        """Write the .obj file.

        :param fname: A filename for the .obj file.
        :param mtllib: The name of a .mtl file to be referenced from the .obj file.
        """
        with open(fname, 'w') as f_out:
            f_out.write('# exported using geomeppy\n# https://github.com/jamiebull1/geomeppy\n')
            f_out.write('\n# vertices\n')
            for v in self.vertices:
                f_out.write('v %.6f %.6f %.6f\n' % (v[0], v[1], v[2]))
            f_out.write('\n# materials library\n')
            f_out.write('mtllib %s\n' % os.path.basename(mtllib))
            f_out.write('\n# faces\n')
            for f in self.faces:
                f_out.write('usemtl %s\n' % f['mtl'].lower())
                f_out.write('f %s//\n' % '// '.join((str(i) for i in f['face'])))


def export_to_obj(idf, fname=None, mtllib=None):
    # type: (IDF, Optional[str], Optional[str]) -> None
    """Export an OBJ file representation of the IDF.

    This can be used for viewing in tools which support the .obj format.

    :param idf: An IDF to export.
    :param fname: A filename for the .obj file. If None we try to base it on IDF.idfname and change the filetype.
    :param mtllib: The name of a .mtl file to be referenced from the .obj file. If None, we use default.mtl.
    """
    if not fname:
        name, _ext = os.path.splitext(idf.idfname)
        fname = '%s.obj' % name
    if not mtllib:
        name, _ext = os.path.splitext(fname)
        mtllib = '%s.mtl' % name
        shutil.copy(os.path.join(THIS_DIR, 'default.mtl'), mtllib)
    obj_writer = ObjWriter()
    surfaces = idf.getsurfaces()
    subsurfaces = idf.getsubsurfaces()
    shading_surfaces = idf.getshadingsurfaces()
    obj_writer.from_surfaces(surfaces, subsurfaces, shading_surfaces)
    obj_writer.write(fname, mtllib)



"""
A module to handle translations, using Christopher Gohlke's transforms3d as far
as possible, but also trying to respect the intent of the algorithms used in
OpenStudio for the sake of consistency between tools based on EnergyPlus.

"""

from typing import Any, Optional, Union  # noqa

import numpy as np
from transforms3d._gohlketransforms import (
    concatenate_matrices,
    identity_matrix,
    inverse_matrix,
    rotation_matrix,
    translation_matrix,
)

if False: from .polygons import Polygon3D  # noqa
from .vectors import Vector2D, Vector3D  # noqa


class Transformation(object):
    def __init__(self, mat=None):
        # type: (Optional[np.ndarray]) -> None
        if mat is None:
            # initialise with a 4D identity matrix
            self.matrix = identity_matrix()
        else:
            # initialise with the matrix passed in
            self.matrix = mat
        assert self.matrix.shape == (4, 4)

    def __mul__(self, other):
        # type: (Any) -> Union[Transformation, Vector3D]
        if hasattr(other, 'matrix'):
            # matrix by a matrix
            mat = concatenate_matrices(self.matrix, other.matrix)  # type: ignore
            return Transformation(mat)
        elif hasattr(other, 'x'):
            # matrix by a vector
            vector = other
            temp = [vector.x, vector.y, vector.z, 1]  # type: ignore
            result = np.dot(self.matrix, temp)[:3]
            return Vector3D(*result)
        else:
            # matrix by each point in a polygon
            result = [self * point for point in other]
            return other.__class__(result)

    def _align_z_prime(self, zp):
        # type: (Union[Vector2D, Vector3D]) -> Transformation
        """Transform system with z' to regular system. Will try to align y'
        with z, but if that fails will align y' with y

        """
        zp = zp.normalize()

        z_axis = Vector3D(0, 0, 1)
        neg_x_axis = Vector3D(-1, 0, 0)

        # check if face normal is up or down
        if abs(zp.dot(z_axis)) < 0.99:
            # not facing up or down, set yPrime along z_axis
            yp = z_axis - zp.dot(z_axis) * zp
            yp = yp.normalize()
            xp = yp.cross(zp)
        else:
            # facing up or down, set xPrime along -x_axis
            xp = neg_x_axis - zp.dot(neg_x_axis) * zp
            xp = xp.normalize()
            yp = zp.cross(xp)

        self.matrix[:3, 0] = xp
        self.matrix[:3, 1] = yp
        self.matrix[:3, 2] = zp

        return self

    def _align_face(self, polygon):
        """A transformation to align a polygon with the origin.

        Parameters
        ----------
        polygon : Polygon3D

        Returns
        -------
        Transformation

        """
        zp = polygon.normal_vector
        align = self._align_z_prime(zp)

        aligned_vertices = align._inverse() * polygon
        min_x = min(aligned_vertices.xs)
        min_y = min(aligned_vertices.ys)
        min_z = min(aligned_vertices.zs)
        assert aligned_vertices.is_horizontal

        direction = Vector3D(min_x, min_y, min_z)
        translate = self._translation(direction)

        self.matrix = concatenate_matrices(align.matrix, translate.matrix)

        return self

    def _inverse(self):
        # type: () -> Transformation
        """The inverse of a given transformation.

        Returns
        -------
        Transformation

        """
        return Transformation(inverse_matrix(self.matrix))

    def _translation(self, direction):
        # type: (Vector3D) -> Transformation
        return Transformation(translation_matrix(direction))

    def _rotation(self, direction, angle):
        # type: (Vector3D, Union[int, np.float]) -> Transformation
        return Transformation(rotation_matrix(angle, direction))


def align_face(polygon):
    """Transformation to align face with z-axis.

    :param polygon: Polygon to be aligned.
    :returns: Polygon3D aligned with the z-axis.
    """
    t = Transformation()
    t._align_face(polygon)

    return t._inverse() * polygon


def invert_align_face(original, poly2):
    """Transformation to align face with original position.

    :param original: Polygon in the desired orientation.
    :param poly2: Polygon previously aligned with `align_face`.
    :returns: Polygon returned to the original orientation.
    """
    t = Transformation()
    t._align_face(original)

    return t * poly2



"""
Perform clipping operations on Polygons
---------------------------------------

PyClipper is used for clipping. It's a wrapper for the C++ version of the Clipper library.

We implement a few of the functions of PyClipper here as `.difference`, `.intersect`, and `.union` methods of the
`Clipper2D` and `Clipper3D` classes. These are then used as mixins for the `Polygon2D` and `Polygon3D` classes.

"""
from typing import List  # noqa

import pyclipper as pc

if False: from .polygons import Polygon  # noqa
from ..utilities import almostequal


class Clipper2D(object):
    """This class is used to add clipping functionality to the Polygon2D class."""

    def difference(self, poly):
        # type: (Polygon) -> List[Polygon]
        """Difference from another polygon.

        :param poly: The clip polygon.
        :returns: A list of Polygons representing the difference.

        """
        clipper = self._prepare_clipper(poly)
        if not clipper:
            return []
        differences = clipper.Execute(
            pc.CT_DIFFERENCE, pc.PFT_NONZERO, pc.PFT_NONZERO)

        return self._process(differences)

    def intersect(self, poly):
        # type: (Polygon) -> List[Polygon]
        """Intersect with another polygon.

        :param poly: The clip polygon.
        :returns: False if no intersection, otherwise a list of Polygons representing each intersection.

        """
        clipper = self._prepare_clipper(poly)
        if not clipper:
            return []
        intersections = clipper.Execute(
            pc.CT_INTERSECTION, pc.PFT_NONZERO, pc.PFT_NONZERO)

        return self._process(intersections)

    def union(self, poly):
        # type: (Polygon) -> List[Polygon]
        """Union with another polygon.

        :param poly: The clip polygon.
        :returns: A list of Polygons.

        """
        clipper = self._prepare_clipper(poly)
        if not clipper:
            return []
        unions = clipper.Execute(
            pc.CT_UNION, pc.PFT_NONZERO, pc.PFT_NONZERO)

        return self._process(unions)

    def _prepare_clipper(self, poly):
        """Prepare 2D polygons for clipping operations.

        :param poly: The clip polygon.
        :returns: A Pyclipper object.

        """
        s1 = pc.scale_to_clipper(self.vertices_list)
        s2 = pc.scale_to_clipper(poly.vertices_list)
        clipper = pc.Pyclipper()
        clipper.AddPath(s1, poly_type=pc.PT_SUBJECT, closed=True)
        clipper.AddPath(s2, poly_type=pc.PT_CLIP, closed=True)
        return clipper

    def _process(self, results):
        """Process and return the results of a clipping operation.

        :param results: A list of lists of coordinates .
        :returns: A list of Polygon2D results of the clipping operation.

        """
        if not results:
            return []
        scaled = [pc.scale_from_clipper(r) for r in results]
        polys = [self.as_2d(r) for r in scaled]
        processed = []
        for poly in polys:
            if almostequal(poly.normal_vector, self.normal_vector):
                processed.append(poly)
            else:
                processed.append(poly.invert_orientation())
        return processed


class Clipper3D(Clipper2D):
    """This class is used to add clipping functionality to the Polygon3D class."""

    def _prepare_clipper(self, poly):
        """Prepare 3D polygons for clipping operations.

        :param poly: The clip polygon.
        :returns: A Pyclipper object.

        """
        if not self.is_coplanar(poly):
            return False
        poly1 = self.project_to_2D()
        poly2 = poly.project_to_2D()

        s1 = pc.scale_to_clipper(poly1.vertices_list)
        s2 = pc.scale_to_clipper(poly2.vertices_list)
        clipper = pc.Pyclipper()
        clipper.AddPath(s1, poly_type=pc.PT_SUBJECT, closed=True)
        clipper.AddPath(s2, poly_type=pc.PT_CLIP, closed=True)

        return clipper

    def _process(self, results):
        """Process and return the results of a clipping operation.

        :param results: A list of lists of coordinates .
        :returns: A list of Polygon3D results of the clipping operation.

        """
        if not results:
            return []
        results = [pc.scale_from_clipper(r) for r in results]
        polys = [self.as_2d(v).project_to_3D(self) for v in results]
        processed = []
        for poly in polys:
            if almostequal(self.normal_vector, poly.normal_vector):
                processed.append(poly)
            else:
                processed.append(poly.invert_orientation())
        return processed



"""A collection of functions which act on surfaces or lists of surfaces."""
from collections import defaultdict
from itertools import combinations
from typing import Dict, List, Tuple, Union  # noqa
import warnings

from eppy.bunch_subclass import EpBunch  # noqa
from eppy.idf_msequence import Idf_MSequence  # noqa
from numpy import float64  # noqa

from .polygons import intersect, Polygon3D, unique
from .vectors import Vector2D, Vector3D  # noqa
from ..utilities import almostequal


def set_coords(surface,  # type: EpBunch
               coords,  # type: Union[List[Vector3D], List[Tuple[float, float, float]], Polygon3D]
               ggr  # type: Union[List, None, Idf_MSequence]
               ):
    # type: (...) -> None
    """Update the coordinates of a surface.

    :param surface: The surface to modify.
    :param coords: The new coordinates as lists of [x,y,z] lists.
    :param ggr: Global geometry rules.
    """
    poly = Polygon3D(coords)
    poly = poly.normalize_coords(ggr)
    coords = [i for vertex in poly for i in vertex]
    if len(coords) > 120:
        warnings.warn(
            'To create surfaces with >120 vertices, ensure you have customised your IDD before running EnergyPlus. '
            'https://unmethours.com/question/9343/energy-idf-parsing-error/?answer=9344#post-id-9344'
        )
    # find the vertex fields
    n_vertices_index = surface.objls.index('Number_of_Vertices')
    first_x = n_vertices_index + 1  # X of first coordinate
    surface.obj = surface.obj[:first_x]
    # set the vertex field values
    surface.fieldvalues.extend(coords)


def set_matched_surfaces(surface, matched):
    # type: (EpBunch, EpBunch) -> None
    """Set boundary conditions for two adjoining surfaces.

    :param surface: The first surface.
    :param matched: The second surface.
    """
    for s in [surface, matched]:
        s.Outside_Boundary_Condition = 'surface'
        s.Sun_Exposure = 'NoSun'
        s.Wind_Exposure = 'NoWind'
    surface.Outside_Boundary_Condition_Object = matched.Name
    matched.Outside_Boundary_Condition_Object = surface.Name


def set_unmatched_surface(surface, vector):
    # type: (EpBunch, Union[Vector2D, Vector3D]) -> None
    """Set boundary conditions for a surface which does not adjoin another one.

    :param surface: The surface.
    :param vector: The surface normal vector.
    """
    surface.View_Factor_to_Ground = 'autocalculate'
    poly = Polygon3D(surface.coords)
    if min(poly.zs) < 0 or all(z == 0 for z in poly.zs):
        # below ground or ground-adjacent surfaces
        surface.Outside_Boundary_Condition_Object = ''
        surface.Outside_Boundary_Condition = 'ground'
        surface.Sun_Exposure = 'NoSun'
        surface.Wind_Exposure = 'NoWind'
    else:
        surface.Outside_Boundary_Condition = 'outdoors'
        surface.Outside_Boundary_Condition_Object = ''
        surface.Wind_Exposure = 'WindExposed'
        if almostequal(vector, (0, 0, -1)):
            # downward facing surfaces
            surface.Sun_Exposure = 'NoSun'
        else:
            surface.Sun_Exposure = 'SunExposed'  # other external surfaces


def getidfplanes(surfaces):
    # type: (Idf_MSequence) -> Dict[float64, Dict[Union[Vector2D, Vector3D], List[EpBunch]]]
    """Fast access data structure for potentially matched surfaces.

    Get a data structure populated with all the surfaces in the IDF, keyed by their distance from the origin, and their
    normal vector.

    :param surfaces: List of all the surfaces.
    :returns: Mapping to look up IDF surfaces.
    """
    planes = {}  # type: Dict[float64, Dict[Union[Vector2D, Vector3D], List[EpBunch]]]
    for s in surfaces:
        poly = Polygon3D(s.coords)
        rounded_distance = round(poly.distance, 8)
        rounded_normal_vector = Vector3D(*[round(axis, 8)
                                           for axis in poly.normal_vector])
        planes.setdefault(rounded_distance,
                          {}).setdefault(rounded_normal_vector,
                                         []).append(s)
    return planes


def get_adjacencies(surfaces):
    # type: (Idf_MSequence) -> defaultdict
    """Create a dictionary mapping surfaces to their adjacent surfaces.

    :param surfaces: A mutable list of surfaces.
    :returns: Mapping of surfaces to adjacent surfaces.
    """
    adjacencies = defaultdict(list)  # type: defaultdict
    # find all adjacent surfaces
    for s1, s2 in combinations(surfaces, 2):
        adjacencies = populate_adjacencies(adjacencies, s1, s2)
    # make sure we have only unique surfaces
    for surface in adjacencies:
        adjacencies[surface] = unique(adjacencies[surface])

    return adjacencies


def populate_adjacencies(adjacencies, s1, s2):
    # type: (defaultdict, EpBunch, EpBunch) -> defaultdict
    """Update the adjacencies dict with any intersections between two surfaces.

    :param adjacencies: Dict to contain lists of adjacent surfaces.
    :param s1: Object representing an EnergyPlus surface.
    :param s2: Object representing an EnergyPlus surface.
    :returns: An updated dict of adjacencies.
    """
    poly1 = Polygon3D(s1.coords)
    poly2 = Polygon3D(s2.coords)
    if not almostequal(abs(poly1.distance), abs(poly2.distance), 4):
        return adjacencies
    if not almostequal(poly1.normal_vector, poly2.normal_vector, 4):
        if not almostequal(poly1.normal_vector, -poly2.normal_vector, 4):
            return adjacencies

    intersection = poly1.intersect(poly2)
    if intersection:
        new_surfaces = intersect(poly1, poly2)
        new_s1 = [s for s in new_surfaces
                  if almostequal(s.normal_vector, poly1.normal_vector, 4)]
        new_s2 = [s for s in new_surfaces
                  if almostequal(s.normal_vector, poly2.normal_vector, 4)]
        adjacencies[(s1.key, s1.Name)] += new_s1
        adjacencies[(s2.key, s2.Name)] += new_s2
    return adjacencies



import os

from setuptools import setup


THIS_DIR = os.path.dirname(os.path.abspath(__file__))

def read_rst(f):
    try:
        with open(os.path.join(THIS_DIR, f), 'r') as f_in:
            return f_in.read()
    except:
        return "GeomEppy"

setup(
    name='geomeppy',
    packages=[
        'geomeppy',
        'geomeppy.geom',
        'geomeppy.io',
        'tests',
    ],
    version='0.4.8',
    description='Geometry editing for E+ idf files',
    long_description=read_rst('README.rst'),
    author='Jamie Bull',
    author_email='jamie.bull@oco-carbon.com',
    url='https://github.com/jamiebull1/geomeppy',
    download_url='https://github.com/jamiebull1/geomeppy/tarball/v0.4.8',
    license='MIT License',
    keywords=[
        'EnergyPlus',
        'geometry',
        'building performance simulation',
    ],
    platforms='any',
    install_requires=[
        'eppy==0.5.46',
        'matplotlib==2.1.1',  # model viewer
        'numpy==1.14',
        'pyclipper==1.1.0',  # geometry intersection
        'shapely==1.6.2',  # geometry transformations
        'six==1.11.0',  # python2/3 compatibility
        'transforms3d==0.3.1',  # geometry transformations
        'pypoly2tri==0.0.3',  # triangulate polygons
    ],
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
    ],
    extras_require={
        ':python_version>="3.4"': [
            'mypy==0.560',  # static type checking
        ],
        'testing': [
            'codecov',
            'flake8',
            'pytest-cov',
            'typing',
        ],
    },
    include_package_data=True,
)




"""Intersect and match all surfaces in an IDF.
"""
from typing import List, Optional, Union  # noqa

from eppy.idf_msequence import Idf_MSequence  # noqa

from geomeppy.geom.polygons import Polygon3D
from geomeppy.geom.surfaces import (
    get_adjacencies, getidfplanes, set_coords, set_matched_surfaces, set_unmatched_surface,
)
from geomeppy.utilities import almostequal

if False: from ..idf import IDF  # noqa


def intersect_idf_surfaces(idf):
    # type: (IDF) -> None
    """Intersect all surfaces in an IDF.

    :param idf: The IDF.
    """
    surfaces = idf.getsurfaces()
    try:
        ggr = idf.idfobjects['GLOBALGEOMETRYRULES'][0]
    except IndexError:
        ggr = None
    # get all the intersected surfaces
    adjacencies = get_adjacencies(surfaces)
    for surface in adjacencies:
        key, name = surface
        new_surfaces = adjacencies[surface]
        old_obj = idf.getobject(key.upper(), name)
        for i, new_coords in enumerate(new_surfaces, 1):
            new = idf.copyidfobject(old_obj)
            new.Name = "%s_%i" % (name, i)
            set_coords(new, new_coords, ggr)
        idf.removeidfobject(old_obj)


def match_idf_surfaces(idf):
    # type: (IDF) -> None
    """Match all surfaces in an IDF.

    :param idf: The IDF.
    """
    surfaces = idf.getsurfaces()
    planes = getidfplanes(surfaces)
    for distance in planes:
        for vector in planes[distance]:
            surfaces = planes[distance][vector]
            matches = planes.get(-distance, {}).get(-vector, [])
            if not matches:
                # default set all the surfaces boundary conditions
                for s in surfaces:
                    set_unmatched_surface(s, vector)
            else:
                # check which are matches
                for s in surfaces:
                    for m in matches:
                        matched = False
                        poly1 = Polygon3D(s.coords)
                        poly2 = Polygon3D(m.coords)
                        if almostequal(poly1, poly2.invert_orientation()):
                            matched = True
                            set_matched_surfaces(s, m)
                            break
                    if not matched:
                        # unmatched surfaces
                        set_unmatched_surface(s, vector)
                        set_unmatched_surface(m, vector)
